/*
    UHU-menu -- input plugin for freedesktop-style files
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhumenu.h"

static struct menu *menu_new (void)
{
	struct menu *menu;
	menu = malloc(sizeof(struct menu));
	menu->nr = 0;
	menu->items = NULL;
	return menu;
}

static int menu_add_item (struct menu *menu, const struct item *item)
{
	if (menu->nr == 0) {
		menu->items = malloc(sizeof(struct item));
	} else {
		menu->items = realloc(menu->items, (menu->nr + 1) * sizeof(struct item));
	}
	if (item != NULL) {
		memcpy(&menu->items[menu->nr], item, sizeof(struct item));
	}
	menu->nr++;
	return 0;
}

static int replace_item (struct item *dst, struct item *src)
{
	src->issubmenu = dst->issubmenu;
	src->submenu = dst->submenu;
	src->file_name = dst->file_name;
	memcpy(dst, src, sizeof(struct item));
	return 0;
}

static int menu_insert_directory_item (struct menu *menu, struct item *item,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	path += strspn(path, "/");

	if (path[0] == '\0') {
		if (menu->nr > 0 && menu->items[0].sections != NULL) {
			message(3, "root menu directory duplicated");
			return -1;
		}
		replace_item(&menu->items[0], item);
		return 0;
	}
	buf[0] = '\0';
	while (1) {
		for (i = 0; i < menu->nr; i++) {
			if (!(strcmp(menu->items[i].file_name, buf))) {
				break;
			}
		}
		if (i == menu->nr) {
			menu_add_item(menu, NULL);
			menu->items[i].issubmenu = 1;
			menu->items[i].submenu = menu_new();
			menu->items[i].file_name = strdup(buf);
			menu->items[i].sections = NULL;
		}
		if (path[0] == '\0') break;
		menu = menu->items[i].submenu;
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
	}
	replace_item(&menu->items[i], item);
	return 0;
}

static int menu_insert_file_item (struct menu *menu, struct item *item,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	path += strspn(path, "/");

	buf[0] = '\0';
	while (1) {
		for (i = 0; i < menu->nr; i++) {
			if (!(strcmp(menu->items[i].file_name, buf))) {
				break;
			}
		}
		if (i == menu->nr) {
			menu_add_item(menu, NULL);
			menu->items[i].issubmenu = 0;
			menu->items[i].submenu = NULL;
			menu->items[i].file_name = strdup(buf);
			menu->items[i].sections = NULL;
		}
		if (path[0] == '\0') break;
		menu = menu->items[i].submenu;
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
	}
	replace_item(&menu->items[i], item);
	return 0;
}

struct menu *read_menu (const char *path)
{
	struct menu *menu;
	struct item it;
	char *fn;
	char fn2[PATH_MAX];
	char buf[PATH_MAX];
	char buf2[PATH_MAX];
	struct dirent *de;
	DIR *d;
	int l;

	menu = menu_new();
	it.issubmenu = 1;

	sprintf(buf, "%s/desktop-directories", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		l = strlen(de->d_name);
		if (l < 10 || strcmp(de->d_name + l - 10, ".directory")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
		if (desktop_parse(buf2, 0, &it) >= 0) {
			fn = get_value(&it, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				menu_insert_directory_item(menu, &it, fn);
			}
		}
	}
	closedir(d);

	sprintf(buf, "%s/applications", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		l = strlen(de->d_name);
		if (l < 8 || strcmp(de->d_name + l - 8, ".desktop")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
		if (desktop_parse(buf2, 0, &it) >= 0) {
			fn = get_value(&it, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				sprintf(fn2, "%s/%s", fn, de->d_name);
				menu_insert_file_item(menu, &it, fn2);
			}
		}
	}
	closedir(d);

	return menu;
}
