/*
    UHU-menu
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <dlfcn.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>

#define UHU_MENU_CORE_ENGINE
#include "uhumenu.h"

#define MENU_SRC_DIR "/usr/lib"
#define MENU_DST_DIR "/usr/lib/menu"
#define PLUGIN_DIR "/usr/lib/uhumenu"
#define INPUT_PLUGIN "desktop"

const char *menu_src_dir = MENU_SRC_DIR;
const char *menu_dst_dir = MENU_DST_DIR;
const char *plugin_dir = PLUGIN_DIR;
const char *input_plugin = INPUT_PLUGIN;
char input_plugin_dir[PATH_MAX];
char output_plugin_dir[PATH_MAX];

#define LOGLEVEL 5
int loglevel = LOGLEVEL;

char buf[PATH_MAX];

char *wm_only = NULL;
char *wm_skip = NULL;

char *lang_only = NULL;
char *lang_add = NULL;
char *lang_skip = NULL;

struct language *current_language;

int plugin_type = -1;
const char *plugin_name = NULL;

#define PLUGINS_MAX 64
struct plugin {
	char *name;
	int type;
	int ok;
	void *handle;
	struct menu *(*read_menu) (const char *path);
	int (*write_menu) (const struct menu *menu, const char *path);
	int (*write_menu_lang) (const struct menu *menu, const char *path, const struct language *language);
};

struct plugin iplugin;
struct plugin oplugins[PLUGINS_MAX];
int opluginnr = 0;

int plugin_sort_fn (const void *v1, const void *v2)
{
	const struct plugin *p1 = v1;
	const struct plugin *p2 = v2;
	return strcmp(p1->name, p2->name);
}

/*
 *  qsort_menu + menu_sort_fn
 *  Sorts the menu according to current_lang. Invalid entries are put to the
 *  end of the list. "nr" is set to the number of valid entries, as opposed to
 *  nr_real which counts the invalid entries too.
 */
int menu_sort_fn (const void *v1, const void *v2)
{
	const struct item *i1 = v1;
	const struct item *i2 = v2;
	const struct key *k1, *k2;
	if (i1->valid != i2->valid) return i2->valid - i1->valid;
	if (i1->issubmenu != i2->issubmenu) return i2->issubmenu - i1->issubmenu;
	if (i1->major != i2->major) return i2->major - i1->major;
	if (i1->minor != i2->minor) return i2->minor - i1->minor;
	k1 = item_find_key(i1, NULL, "Name");
	k2 = item_find_key(i2, NULL, "Name");
	if (k1 == NULL && k2 == NULL) return 0;
	if (k1 == NULL) return -1;
	if (k2 == NULL) return 1;
	if (k1->current_lang == NULL && k2->current_lang == NULL) return 0;
	if (k1->current_lang == NULL) return -1;
	if (k2->current_lang == NULL) return 1;
	if (current_language->locale_utf8 != NULL) {
		return strcoll_l(k1->current_lang->value, k2->current_lang->value, current_language->locale_utf8);
	} else if (current_language->locale != NULL) {
		return strcoll_l(k1->current_lang->value_8bit, k2->current_lang->value_8bit, current_language->locale);
	} else {
		return strcmp(k1->current_lang->value, k2->current_lang->value);
	}
}

void qsort_menu (struct menu *menu)
{
	int i;
	qsort(menu->items, menu->nr_real, sizeof(struct item), menu_sort_fn);
	for (i = 0; i < menu->nr_real && menu->items[i].valid; i++) {
		if (menu->items[i].issubmenu) {
			qsort_menu(menu->items[i].submenu);
		}
	}
	menu->nr = i;
}

/*
 *  empty_is_invalid
 *  Marks empty submenus (those which do not have valid and nonempty descendants)
 *  as invalid.
 */
int empty_is_invalid (struct menu *menu)
{
	int retval = -1;
	int i;
	for (i = 0; i < menu->nr_real; i++) {
		if (!menu->items[i].valid) continue;
		if (menu->items[i].issubmenu && empty_is_invalid(menu->items[i].submenu) < 0) {
			menu->items[i].valid = 0;
		} else {
			retval = 0;
		}
	}
	return retval;
}

/*
 *  prepare_for_wm
 *  Prepares the menu for a particular window manager. Marks entries that
 *  should be skipped for this wm as invalid.
 */
int prepare_for_wm (struct menu *menu, const char *wm)
{
	int i;
	for (i = 0; i < menu->nr_real; i++) {
		menu->items[i].valid_wm = 1;
		if (menu->items[i].onlyshowin != NULL &&
		  !is_in_list(wm, menu->items[i].onlyshowin, ",;")) {
			menu->items[i].valid_wm = 0;
		}
		if (menu->items[i].notshowin != NULL &&
		  is_in_list(wm, menu->items[i].notshowin, ",;")) {
			menu->items[i].valid_wm = 0;
		}
		menu->items[i].valid = menu->items[i].valid_general && menu->items[i].valid_wm && menu->items[i].valid_lang;
		if (menu->items[i].issubmenu) {
			prepare_for_wm(menu->items[i].submenu, wm);
		}
	}
	return 0;
}

/*
 *  prepare_for_lang
 *  Prepares the menu for a particular language. Sets the current_lang pointer
 *  in every key of every item for faster access. Converts major and minor
 *  entries from string to int. Does not yet sort the entries.
 */
int prepare_for_lang (struct menu *menu, const struct language *language)
{
	int i;
	struct section *s;
	struct key *k;
	for (i = 0; i < menu->nr_real; i++) {
		s=menu->items[i].sections;
		while (s != NULL) {
			k = s->keys;
			while (k != NULL) {
				k->current_lang = key_find_lang(k, language->name);
				k = k->next;
			}
			s = s->next;
		}
		menu->items[i].valid_lang = 1;
		if (menu->items[i].key_name == NULL || menu->items[i].key_name->current_lang == NULL) {
			menu->items[i].valid_lang = 0;
		}
		if (menu->items[i].key_major == NULL ||
		  menu->items[i].key_major->current_lang == NULL) {
			menu->items[i].major = 0;
		} else {
			menu->items[i].major = string_to_int(menu->items[i].key_major->current_lang->value, 0);
		}
		if (menu->items[i].key_minor == NULL ||
		  menu->items[i].key_minor->current_lang == NULL) {
			menu->items[i].minor = 0;
		} else {
			menu->items[i].minor = string_to_int(menu->items[i].key_minor->current_lang->value, 0);
		}
		menu->items[i].valid = menu->items[i].valid_general && menu->items[i].valid_wm && menu->items[i].valid_lang;
		if (menu->items[i].issubmenu) {
			prepare_for_lang(menu->items[i].submenu, language);
		}
	}
	return 0;
}

/*
 *  fix_menu
 *  Called only once after the input plugin has built up the menu tree and
 *  before the input oplugins dump it. Fills out some fields which might yet
 *  be uninitialized but may be needed by this core engine or an output plugin.
 */
int fix_menu (struct menu *menu)
{
	int i;
	struct section *s;
	struct key *k;
	struct lang *l;
	menu->nr_real = menu->nr;
	for (i = 0; i < menu->nr; i++) {
		menu->items[i].key_name  = item_find_key(&menu->items[i], "", "Name");
		menu->items[i].key_major = item_find_key(&menu->items[i], "", "X-UHU-Major");
		menu->items[i].key_minor = item_find_key(&menu->items[i], "", "X-UHU-Minor");
		s = menu->items[i].sections;
		while (s != NULL) {
			k = s->keys;
			while (k != NULL) {
				l = k->langs;
				while (l != NULL) {
					if (l->language->charset != NULL) {
						if (l->value == NULL) {
							if (convert(l->language->charset->to_utf8, l->value_8bit, &l->value) < 0) {
								message(5, "Failed to convert \"%s\" from %s to UTF-8", l->value_8bit, l->language->charset->name);
							}
						} else {
							if (convert(l->language->charset->from_utf8, l->value, &l->value_8bit) < 0) {
								message(5, "Failed to convert \"%s\" from UTF-8 to %s", l->value, l->language->charset->name);
							}
						}
					}
					l = l->next;
				}
				k = k->next;
			}
			s = s->next;
		}
		menu->items[i].valid_general =
		  (get_value_bool(&menu->items[i], NULL, "Hidden", NULL) != 1 &&
		   get_value_bool(&menu->items[i], NULL, "NoDisplay", NULL) != 1);
		menu->items[i].terminal = (get_value_bool(&menu->items[i], NULL, "Terminal", NULL) == 1);
		menu->items[i].onlyshowin = get_value(&menu->items[i], NULL, "OnlyShowIn", NULL);
		menu->items[i].notshowin = get_value(&menu->items[i], NULL, "NotShowIn", NULL);
		if (menu->items[i].issubmenu) {
			fix_menu(menu->items[i].submenu);
		}
	}
	return 0;
}

int output_all_in_one (struct menu *menu)
{
	int i, j;

	prepare_for_lang(menu, current_language);

	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		if (oplugins[i].write_menu == NULL) continue;

		prepare_for_wm(menu, oplugins[i].name);
		empty_is_invalid(menu);
		qsort_menu(menu);

		sprintf(buf, "%s/%s", menu_dst_dir, oplugins[i].name);
		mkdir(buf, 0777);

		message(6, "Invoking write_menu() of output plugin `%s'", oplugins[i].name);
		plugin_type = 1;
		plugin_name = oplugins[i].name;
		j = (*oplugins[i].write_menu)(menu, buf);
		plugin_type = -1;
		plugin_name = NULL;
		if (j >= 0) message(6, "write_menu() returned successfully");
		else message(6, "write_menu() returned an error");
	}
	return 0;
}

int output_one_lang (struct menu *menu)
{
	int i, j;

	if (current_language->locale == NULL) return -1;

	prepare_for_lang(menu, current_language);
	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		if (oplugins[i].write_menu_lang == NULL) continue;

		prepare_for_wm(menu, oplugins[i].name);
		empty_is_invalid(menu);
		qsort_menu(menu);

		sprintf(buf, "%s/%s", menu_dst_dir, oplugins[i].name);
		mkdir(buf, 0777);

		message(6, "Invoking write_menu_lang() of output plugin `%s' for language `%s'", oplugins[i].name, current_language->name);
		plugin_type = 1;
		plugin_name = oplugins[i].name;
		j = (*oplugins[i].write_menu_lang)(menu, buf, current_language);
		plugin_type = -1;
		plugin_name = NULL;
		if (j >= 0) message(6, "write_menu_lang() returned successfully");
		else message(6, "write_menu_lang() returned an error");
	}
	return 0;
}

void list ()
{
	int i;
	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		printf("%-15s (%s)\n",
		  oplugins[i].name,
		  oplugins[i].write_menu_lang != NULL ? "separate config for each language"
		    : "one config for all languages");
	}
}

void help ()
{
	printf(
	  "Usage: uhumenu [options...]\n"
	  "\n"
	  "  -s, --source-dir       Source dir of menu (default: " MENU_SRC_DIR ")\n"
	  "  -d, --destination-dir  Destination dir of menu (default: " MENU_DST_DIR ")\n"
	  "  -p, --plugin-dir       Plugin dir (default: " PLUGIN_DIR ")\n"
	  "  -i, --input-plugin     Input plugin (default: " INPUT_PLUGIN ")\n"
	  "      --list-plugins     List available window manager plugins\n"
	  "      --wm-only          Process only these window managers (comma separated)\n"
	  "      --wm-skip          Skip these window managers (comma separated)\n"
	  "      --lang-only        Process only these languages (comma separated)\n"
	  "      --lang-add         Make sure to process these languages (comma separated)\n"
	  "      --lang-skip        Skip these languages (comma separated)\n"
	  "  -l, --loglevel         Loglevel (0 - 8) (default: %d)\n"
	  "  -h, --help             Guess what :-)\n"
	, LOGLEVEL);
}

int main (int argc, char *argv[])
{
	struct menu *menu;
	const char *err1, *err2;
	DIR *d;
	struct dirent *de;
	int c;
	int l;
	int i;
	int opt_list = 0;
	char *s;

	static const struct option long_options[] = {
		{ "source-dir",      1, NULL, 's' },
		{ "src-dir",         1, NULL, 's' },
		{ "destination-dir", 1, NULL, 'd' },
		{ "dest-dir",        1, NULL, 'd' },
		{ "dst-dir",         1, NULL, 'd' },
		{ "plugin-dir",      1, NULL, 'p' },
		{ "input-plugin",    1, NULL, 'i' },
		{ "list-plugins",    0, NULL, 256 },
		{ "wm-only",         1, NULL, 257 },
		{ "wm-skip",         1, NULL, 258 },
		{ "lang-only",       1, NULL, 259 },
		{ "lang-add",        1, NULL, 260 },
		{ "lang-skip",       1, NULL, 261 },
		{ "loglevel",        1, NULL, 'l' },
		{ "help",            0, NULL, 'h' },
		{ NULL,              0, NULL, 0   }
	};

	/* Initialization */

	setlocale(LC_ALL, "");

	while ((c = getopt_long(argc, argv, "s:d:p:i:l:h", long_options, NULL)) != -1) {
		switch (c) {
		  case 's':
			menu_src_dir = optarg;
			break;
		  case 'd':
			menu_dst_dir = optarg;
			break;
		  case 'p':
			plugin_dir = optarg;
			break;
		  case 'i':
			input_plugin = optarg;
			break;
		  case 256:
			opt_list = 1;
			break;
		  case 257:
		  case 258:
			if (wm_only != NULL || wm_skip != NULL) {
				message(0, "Window manager list already specified");
				exit(1);
			}
			if (c == 257) wm_only = optarg;
			else wm_skip = optarg;
			break;
		  case 259:
			if (lang_only != NULL || lang_add != NULL || lang_skip != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_only = optarg;
			break;
		  case 260:
			if (lang_only != NULL || lang_add != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_add = optarg;
			break;
		  case 261:
			if (lang_only != NULL || lang_skip != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_skip = optarg;
			break;
		  case 'l':
			loglevel = string_to_int(optarg, -1);
			if (loglevel < 0 || loglevel > 8) {
				message(0, "Invalid loglevel");
				exit(1);
			};
			break;
		  case 'h':
			help();
			exit(0);
		  case '?':
		  case ':':
			exit(1);
		}
	}
	sprintf(input_plugin_dir, "%s/plugins/input", plugin_dir);
	sprintf(output_plugin_dir, "%s/plugins/output", plugin_dir);

	find_or_create_charset("UTF-8");
	find_or_create_language("");

	/* Scan output plugin candidates */

	d = opendir(output_plugin_dir);
	if (d == NULL) {
		message_e(0, "Cannot open plugin dir %s", output_plugin_dir);
		exit(1);
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		l = strlen(de->d_name);
		if (l < 4
		  || strcmp(de->d_name + l - 3, ".so")) {
			message(4, "Skipping non-plugin %s/%s", output_plugin_dir, de->d_name);
			continue;
		}
		if (opluginnr == PLUGINS_MAX) {
			message(0, "Too many output plugins found");
			exit(1);
		}
		sprintf(buf, "%.*s", l - 3, de->d_name);
		if (wm_only != NULL && !is_in_list(buf, wm_only, ",;")) continue;
		if (wm_skip != NULL &&  is_in_list(buf, wm_skip, ",;")) continue;
		oplugins[opluginnr].name = strdup(buf);
		oplugins[opluginnr].ok = 1;
		opluginnr++;
	}
	closedir(d);
	qsort(oplugins, opluginnr, sizeof(struct plugin), plugin_sort_fn);

	/* Open output plugins */

	for (i = 0; i < opluginnr; i++) {
		sprintf(buf, "%s/%s.so", output_plugin_dir, oplugins[i].name);
		oplugins[i].handle = dlopen(buf, RTLD_NOW);
		err1 = dlerror();
		if (err1 != NULL) {
			message(3, "dlopen %s: %s", buf, err1);
			continue;
		}
		oplugins[i].write_menu = dlsym(oplugins[i].handle, "write_menu");
		err1 = dlerror();
		oplugins[i].write_menu_lang = dlsym(oplugins[i].handle, "write_menu_lang");
		err2 = dlerror();
		if ((err1 == NULL && err2 == NULL) || (err1 != NULL && err2 != NULL)) {
			oplugins[i].ok = 0;
			message(3, "Ignoring faulty output plugin %s", oplugins[i].name);
			dlclose(oplugins[i].handle);
			continue;
		}
	}

	if (opt_list) {
		list();
		exit(0);
	}

	/* Open input plugin */

	iplugin.name = strdup(input_plugin);
	sprintf(buf, "%s/%s.so", input_plugin_dir, input_plugin);
	iplugin.handle = dlopen(buf, RTLD_NOW);
	err1 = dlerror();
	if (err1 != NULL) {
		message(0, "dlopen %s: %s", buf, err1);
		exit(1);
	}
	iplugin.read_menu = dlsym(iplugin.handle, "read_menu");
	err1 = dlerror();
	if (err1 != NULL) {
		message(0, "Faulty input plugin!");
		exit(1);
	}

	/* Read the menu */

	message(6, "Invoking read_menu() of input plugin `%s'", iplugin.name);
	plugin_type = 0;
	plugin_name = iplugin.name;
	menu = (*iplugin.read_menu)(menu_src_dir);
	plugin_type = -1;
	plugin_name = NULL;
	if (menu == NULL) message(6, "Input plugin returned NULL");
	else message(6, "Input plugin successfully returned");

	if (!menu) {
		message(0, "Could not read menu!");
		exit(1);
	}

	/* Fixing the menu (filling some fields etc.) */

	fix_menu(menu);

	/* Process all-in-one output plugins */

	current_language = find_or_create_language("");
	output_all_in_one(menu);

	/* Process once-for-each-language plugins */

	current_language = find_or_create_language("");
	output_one_lang(menu);

	s = lang_only;
	if (s == NULL) s = lang_add;
	if (s != NULL) {
		char *lang;
		while (*s != '\0') {
			s += strspn(s, ",;");
			if (*s == '\0') break;
			i = strcspn(s, ",;");
			lang = malloc(i + 1);
			strncpy(lang, s, i);
			lang[i] = '\0';
			s += i;
			current_language = find_or_create_language(lang);
			free(lang);
			output_one_lang(menu);
		}
	}

	if (lang_only == NULL) {
		current_language = first_language;
		while (current_language != NULL) {
			if (current_language->name[0] != '\0' &&
			  !is_in_list(current_language->name, lang_add, ",;") &&
			  !is_in_list(current_language->name, lang_skip, ",;"))
				output_one_lang(menu);
			current_language = current_language->next;
		}
	}

	/* Close output plugins and input plugin */

	for (i = 0; i < opluginnr; i++) {
		if (oplugins[i].ok) dlclose(oplugins[i].handle);
	}
	dlclose(iplugin.handle);

	/* Yippie! Survived without segfault :-) */

	return 0;
}
